/*
Copyright (c) 2015-2019 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

module dlib.core.memory;

import std.stdio;
import std.conv;
import std.traits;
import std.datetime;
import std.algorithm;
import core.stdc.stdlib;
import core.exception: onOutOfMemoryError;

import dlib.memory;

/*
 * Tools for manual memory management
 */

private __gshared ulong _allocatedMemory = 0;

ulong allocatedMemory()
{
    return _allocatedMemory;
}

private __gshared Mallocator _defaultGlobalAllocator; 
private __gshared Allocator _globalAllocator;

Allocator globalAllocator()
{
    if (_globalAllocator is null)
    {
        if (_defaultGlobalAllocator is null)
            _defaultGlobalAllocator = Mallocator.instance;
        _globalAllocator = _defaultGlobalAllocator;
    }
    return _globalAllocator;
}

void globalAllocator(Allocator a)
{
    _globalAllocator = a;
}

struct AllocationRecord
{
    string type;
    size_t size;
    string file;
    int line;
    ulong id;
    bool deleted;
}

private
{
    __gshared bool memoryProfilerEnabled = false;
    __gshared AllocationRecord[ulong] records;
    __gshared ulong counter = 0;

    void addRecord(void* p, string type, size_t size, string file = "<undefined>", int line = 0)
    {
        records[cast(ulong)p] = AllocationRecord(type, size, file, line, counter, false);
        counter++;
    }

    void markDeleted(void* p)
    {
        ulong k = cast(ulong)p - psize;
        records[k].deleted = true;
    }
}

void enableMemoryProfiler(bool mode)
{
    memoryProfilerEnabled = mode;
}

void printMemoryLog()
{
    writeln("----------------------------------------------------");
    writeln("               Memory allocation log                ");
    writeln("----------------------------------------------------");
    auto keys = records.keys;
    sort!((a, b) => records[a].id < records[b].id)(keys);
    foreach(k; keys)
    {
        AllocationRecord r = records[k];
        if (r.deleted)
            writefln("         %s - %s byte(s) in %s(%s)", r.type, r.size, r.file, r.line);
        else
            writefln("REMAINS: %s - %s byte(s) in %s(%s)", r.type, r.size, r.file, r.line);
    }
    writeln("----------------------------------------------------");
    writefln("Total amount of allocated memory: %s byte(s)", _allocatedMemory);
    writeln("----------------------------------------------------");
}
    
void printMemoryLeaks()
{
    writeln("----------------------------------------------------");
    writeln("                    Memory leaks                    ");
    writeln("----------------------------------------------------");
    auto keys = records.keys;
    sort!((a, b) => records[a].id < records[b].id)(keys);
    foreach(k; keys)
    {
        AllocationRecord r = records[k];
        if (!r.deleted)
            writefln("%s - %s byte(s) in %s(%s)", r.type, r.size, r.file, r.line);
    }
    writeln("----------------------------------------------------");
    writefln("Total amount of leaked memory: %s byte(s)", _allocatedMemory);
    writeln("----------------------------------------------------");
}

interface Freeable
{
    void free();
}

enum psize = 8;

static if (__VERSION__ >= 2079)
{
    T allocate(T, A...) (A args, string file = __FILE__, int line = __LINE__) if (is(T == class))
    {
        enum size = __traits(classInstanceSize, T);
        void* p = globalAllocator.allocate(size+psize).ptr;
        if (!p)
            onOutOfMemoryError();
        auto memory = p[psize..psize+size];
        *cast(size_t*)p = size;
        _allocatedMemory += size;
        if (memoryProfilerEnabled)
        {
            addRecord(p, T.stringof, size, file, line);
        }
        auto res = emplace!(T, A)(memory, args);
        return res;
    }

    T* allocate(T, A...) (A args, string file = __FILE__, int line = __LINE__) if (is(T == struct))
    {
        enum size = T.sizeof;
        void* p = globalAllocator.allocate(size+psize).ptr;
        if (!p)
            onOutOfMemoryError();
        auto memory = p[psize..psize+size];
        *cast(size_t*)p = size;
        _allocatedMemory += size;
        if (memoryProfilerEnabled)
        {
            addRecord(p, T.stringof, size, file, line);
        }
        return emplace!(T, A)(memory, args);
    }

    T allocate(T) (size_t length, string file = __FILE__, int line = __LINE__) if (isArray!T)
    {
        alias AT = ForeachType!T;
        size_t size = length * AT.sizeof;
        auto mem = globalAllocator.allocate(size+psize).ptr;
        if (!mem)
            onOutOfMemoryError();
        T arr = cast(T)mem[psize..psize+size];
        foreach(ref v; arr)
            v = v.init;
        *cast(size_t*)mem = size;
        _allocatedMemory += size;
        if (memoryProfilerEnabled)
        {
            addRecord(mem, T.stringof, size, file, line);
        }
        return arr;
    }
}
else
{
    T allocate(T, A...) (A args) if (is(T == class))
    {
        enum size = __traits(classInstanceSize, T);
        void* p = globalAllocator.allocate(size+psize).ptr;
        if (!p)
            onOutOfMemoryError();
        auto memory = p[psize..psize+size];
        *cast(size_t*)p = size;
        _allocatedMemory += size;
        if (memoryProfilerEnabled)
        {
            addRecord(p, T.stringof, size);
        }
        auto res = emplace!(T, A)(memory, args);
        return res;
    }

    T* allocate(T, A...) (A args) if (is(T == struct))
    {
        enum size = T.sizeof;
        void* p = globalAllocator.allocate(size+psize).ptr;
        if (!p)
            onOutOfMemoryError();
        auto memory = p[psize..psize+size];
        *cast(size_t*)p = size;
        _allocatedMemory += size;
        if (memoryProfilerEnabled)
        {
            addRecord(p, T.stringof, size);
        }
        return emplace!(T, A)(memory, args);
    }

    T allocate(T) (size_t length) if (isArray!T)
    {
        alias AT = ForeachType!T;
        size_t size = length * AT.sizeof;
        auto mem = globalAllocator.allocate(size+psize).ptr;
        if (!mem)
            onOutOfMemoryError();
        T arr = cast(T)mem[psize..psize+size];
        foreach(ref v; arr)
            v = v.init;
        *cast(size_t*)mem = size;
        _allocatedMemory += size;
        if (memoryProfilerEnabled)
        {
            addRecord(mem, T.stringof, size);
        }
        return arr;
    }
}

void deallocate(T)(ref T obj) if (isArray!T)
{
    void* p = cast(void*)obj.ptr;
    size_t size = *cast(size_t*)(p - psize);
    globalAllocator.deallocate((p - psize)[0..size+psize]);
    _allocatedMemory -= size;
    if (memoryProfilerEnabled)
    {
        markDeleted(p);
    }
    obj.length = 0;
}

void deallocate(T)(T obj) if (is(T == class) || is(T == interface))
{
    Object o = cast(Object)obj;
    void* p = cast(void*)o;
    size_t size = *cast(size_t*)(p - psize);
    destroy(obj);
    globalAllocator.deallocate((p - psize)[0..size+psize]);
    _allocatedMemory -= size;
    if (memoryProfilerEnabled)
    {
        markDeleted(p);
    }
}

void deallocate(T)(T* obj)
{
    void* p = cast(void*)obj;
    size_t size = *cast(size_t*)(p - psize);
    destroy(obj);
    globalAllocator.deallocate((p - psize)[0..size+psize]);
    _allocatedMemory -= size;
    if (memoryProfilerEnabled)
    {
        markDeleted(p);
    }
}

alias New = allocate;
alias Delete = deallocate;
