/*
Copyright (c) 2015-2020 Timur Gafarov

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

module dlib.core.thread;

/*
 * GC-free, Phobos-independent thread module.
 * Inspired by core.thread, with an additional support for thread termination.
 */

version(Windows)
{
    extern(Windows):

    struct SECURITY_ATTRIBUTES
    {
        uint   nLength;
        void*  lpSecurityDescriptor;
        int    bInheritHandle;
    }

    alias LPTHREAD_START_ROUTINE = extern(Windows) uint function(in void*);

    void* CreateThread(
        in SECURITY_ATTRIBUTES*   lpThreadAttributes,
        in size_t                 dwStackSize,
        in LPTHREAD_START_ROUTINE lpStartAddress,
        in void*                  lpParameter,
        in uint                   dwCreationFlags,
        uint*                     lpThreadId
    );

    uint WaitForMultipleObjects(
        in uint nCount,
        in void** lpHandles,
        in int bWaitAll,
        in uint dwMilliseconds
    );

    int TerminateThread(
        void* hThread,
        in uint dwExitCode
    );

    int GetExitCodeThread(
        in void* hThread,
        uint* lpExitCode
    );

    int CloseHandle(
        in void* hObject
    );

    enum INFINITE = uint.max;
    enum STILL_ACTIVE = 259;
}

version(Posix)
{
    import core.sys.posix.pthread;
}

version(Windows)
{
    private extern(Windows) void Sleep(uint msec);
}
else version(Posix)
{
    import core.stdc.time;
    import core.sys.posix.time;
    import core.sys.posix.signal;
}

class Thread
{
    private void function() func;
    private void delegate() dlgt;
    private bool callFunc;
    private bool initialized = false;

    version(Windows)
    {
        private void* winThread;
    }

    version(Posix)
    {
        private pthread_t posixThread;
        private bool running = false;
    }

    this(void function() func)
    {
        this.func = func;
        callFunc = true;
    }

    this(void delegate() dlgt)
    {
        this.dlgt = dlgt;
        callFunc = false;
    }

    ~this()
    {
        version(Windows)
        {
            if (winThread)
                CloseHandle(winThread);
        }

        version(Posix)
        {
            if (initialized)
                pthread_detach(posixThread);
        }
    }

    void start()
    {
        version(Windows)
        {
            if (winThread)
                CloseHandle(winThread);
            uint threadId;
            winThread = CreateThread(null, cast(size_t)0, &winThreadFunc, cast(void*)this, cast(uint)0, &threadId);
            assert(winThread !is null);
            initialized = true;
        }

        version(Posix)
        {
            running = true;
            pthread_create(&posixThread, null, &posixThreadFunc, cast(void*)this);
            // TODO: validate thread creation
            initialized = true;
        }
    }

    void join()
    {
        version(Windows)
        {
            WaitForMultipleObjects(1, &winThread, 1, INFINITE);
        }

        version(Posix)
        {
            pthread_join(posixThread, null);
        }
    }

    bool isRunning()
    {
        version(Windows)
        {
            uint c = 0;
            GetExitCodeThread(winThread, &c);
            return (c == STILL_ACTIVE);
        }

        version(Posix)
        {
            return running;
        }
    }

    void terminate()
    {
        version(Windows)
        {
            TerminateThread(winThread, 1);
        }

        version(Posix)
        {
            pthread_cancel(posixThread);
            running = false;
        }
    }

    version(Windows)
    {
        extern(Windows) static uint winThreadFunc(in void* lpParam)
        {
            Thread t = cast(Thread)lpParam;
            if (t.callFunc)
                t.func();
            else
                t.dlgt();
            return 0;
        }
    }

    version(Posix)
    {
        extern(C) static void* posixThreadFunc(void* arg)
        {
            Thread t = cast(Thread)arg;
            if (t.callFunc)
                t.func();
            else
                t.dlgt();
            t.running = false;
            return null;
        }
    }
    
    version(Windows)
    {
        static void sleep(uint msec)
        {
            Sleep(msec);
        }
    }
    else version(Posix)
    {
        static void sleep(uint msec)
        {
            timespec ts;
            ts.tv_sec = msec / 1000;
            ts.tv_nsec = (msec % 1000) * 1000000;
            nanosleep(&ts, null);
        }
    }
}
